/*____________________________________________________________________________
**
** File:          optiman
** Description:   optimize nroff man pages for terminal or printer
** Author:        Alexander V. Lukyanov
**____________________________________________________________________________
*/
#include <stdio.h>
#include <unistd.h>
#include <memory.h>
#include <ctype.h>
#include <stdlib.h>
#include <locale.h>
#include <wchar.h>
#include <wctype.h>
#include <errno.h>

#ifndef WITHOUT_CURSES
#include <curses.h>
#include <term.h>
#endif

char  Version[]="1.1.0 (Oct 14 2011)";

#define  LAYERSNUM   8
#define  LAYERLEN 256

#define  USECR    0
#define  DELETE   1
#define  TERM     2

#define  UNDERLN  1
#define  ITALIC   2
#define  BOLD     4

FILE  *in;
FILE  *out;

wchar_t  Layer[LAYERSNUM][LAYERLEN];

char  *Program;

int   OptimizeType=USECR;
int   CurrentAttr=0;
int   NewAttr;

int   html=0;

int   Level(wchar_t ch)
{
   if(ch==0)
      return(0);
   if(iswspace(ch))
      return(1);
   if(ch=='_')
      return(2);
   if(iswpunct(ch))
      return(3);
   if(iswalnum(ch))
      return(4);
   return(5);
}

#ifndef WITHOUT_CURSES
int   MyPutc(int ch)
{
   return(fputwc(ch,out));
}

void  SetAttr(void)
{
   if((CurrentAttr&BOLD) && !(NewAttr&BOLD))
   {
      CurrentAttr&=~(BOLD|UNDERLN);
      tputs(exit_attribute_mode,1,MyPutc);
   }
   if((CurrentAttr&UNDERLN) && !(NewAttr&UNDERLN))
   {
      if(exit_underline_mode)
      {
         CurrentAttr&=~UNDERLN;
         tputs(exit_underline_mode,1,MyPutc);
      }
      else
      {
         CurrentAttr&=~(BOLD|UNDERLN);
         tputs(exit_attribute_mode,1,MyPutc);
      }
   }         
   if((CurrentAttr&ITALIC) && !(NewAttr&ITALIC))
   {
         CurrentAttr&=~ITALIC;
         tputs(exit_italics_mode,1,MyPutc);
   }

   if(!(CurrentAttr&UNDERLN) && (NewAttr&UNDERLN))
   {
      CurrentAttr|=UNDERLN;
      tputs(enter_underline_mode,1,MyPutc);
   }
   if(!(CurrentAttr&ITALIC) && (NewAttr&ITALIC))
   {
      CurrentAttr|=ITALIC;
      tputs(enter_italics_mode,1,MyPutc);
   }
   if(!(CurrentAttr&BOLD) && (NewAttr&BOLD))
   {
      CurrentAttr|=BOLD;
      tputs(enter_bold_mode,1,MyPutc);
   }
}
#endif

void  PutOut(void)
{
   int   i;
   int   pos;
   
   switch(OptimizeType)
   {
   case(USECR):
      for(i=0; Layer[i][0] && i<LAYERSNUM; i++)
      {
         fputws(Layer[i],out);
         if(i==LAYERSNUM-1 || Layer[i+1][0]==0)
            break;
         fputwc('\r',out);
      }
      break;
   case(DELETE):
      for(pos=0; pos<LAYERLEN; pos++)
      {
         wchar_t ch=0;
         int   lev=0;
         
         for(i=0; Layer[i][pos] && i<LAYERSNUM; i++)
         {
            if(lev<Level(Layer[i][pos]))
            {
               ch=Layer[i][pos];
               lev=Level(Layer[i][pos]);
            }
         }
         if(ch==0)
            break;
         fputwc(ch,out);
      }
      break;

#ifndef WITHOUT_CURSES
   case(TERM):
      for(pos=0; pos<LAYERLEN; pos++)
      {
         wchar_t ch=0;
         int   underln=0;
         int   bold=0;
         
         for(i=0; Layer[i][pos] && i<LAYERSNUM; i++)
         {
            if(ch==0)
               ch=Layer[i][pos];
            else if(ch==Layer[i][pos])
            {
               if(!isspace(ch) && ch!='_')
                  bold=1;
            }
            else if(ch=='_' && !isspace(Layer[i][pos]))
            {
               underln=1;
               bold=0;
               ch=Layer[i][pos];
            }
            else if(Layer[i][pos]=='_')
               underln=1;
         }
         if(ch==0)
            break;
         NewAttr=0;
         if(underln)
         {
            if(enter_italics_mode && exit_italics_mode)
               NewAttr|=ITALIC;
            else
               NewAttr|=UNDERLN;
         }
         if(bold)
         {
            NewAttr|=BOLD;
         }
/* set new attributes only if ch is not a space
   or underline attr is to be changed,
   because underline alter space and all other attributes don't */
         if(ch!=' ' || html || ((NewAttr^CurrentAttr)&UNDERLN))
            SetAttr();
         if(html)
         {
            switch(ch)
            {
            case '<': fputs("&lt;",out); break;
            case '>': fputs("&gt;",out); break;
            case '&': fputs("&amp;",out); break;
            default: fputwc(ch,out);
            }
         }
         else /* !html */
         {
            fputwc(ch,out);
         }
      }
      if(html)
      {
         NewAttr=0;
         SetAttr();
      }
      break;
#endif /* WITHOUT_CURSES */
   }
}

void  Optimize()
{
   wint_t ch;
   int   pos;
   int   i,j;
   
   pos=0;
   for(i=0; i<LAYERSNUM; i++)
      memset(Layer[i],0,LAYERLEN*sizeof(Layer[i][0]));
   
   while(!feof(in))
   {
      errno=0;
      ch=fgetwc(in);
      if(ch==WEOF)
         break;
      if(ch=='\n')
      {
         PutOut();
         fputwc('\n',out);
         pos=0;
         for(i=0; i<LAYERSNUM; i++)
            memset(Layer[i],0,LAYERLEN*sizeof(Layer[i][0]));
         continue;
      }
      if(ch=='\t')
      {
         pos=(pos|7)+1;
         continue;
      }
      if(ch=='\r')
      {
         pos=0;
         continue;
      }
      if(ch=='\b')
      {
         if(pos>0)
            pos--;
         continue;
      }
      if(pos<LAYERLEN-1)
      {
         for(i=0; i<LAYERSNUM; i++)
         {
            if(Layer[i][pos]==0 || Layer[i][pos]==' ')
            {
               Layer[i][pos]=ch;
               for(j=pos-1; j>=0; j--)
               {
                  if(Layer[i][j]==0)
                     Layer[i][j]=' ';
               }
               break;
            }
         }
      }
      pos++;
      if(ferror(out))
      {
         fprintf(stderr,"%s: Writing error",Program);
         exit(1);
      }
   }
   if(ferror(in) && errno)
   {
      fprintf(stderr,"%s: Reading error (%s)",Program,strerror(errno));
      exit(1);
   }
}

int   main(int argc,char **argv)
{
   int   opt;
   char  *term=0;
   extern int  optind;
   extern char *optarg;

   setlocale(LC_ALL,"");
   
   in=stdin;
   out=stdout;
   
   Program=argv[0];
   
   while((opt=getopt(argc,argv,"dt:h"))!=-1)
   {
      switch(opt)
      {
      case('d'):
         OptimizeType=DELETE;
         break;
      case('t'):
#ifndef WITHOUT_CURSES
         term=optarg;
         OptimizeType=TERM;
#else
         fprintf(stderr,"%s: curses support is not compiled in\n",Program);
         exit(1);
#endif
         break;
      case('?'):
      case('h'):
         printf("Optiman | Version %s | Written by Alexander V. Lukyanov (1994,2001)\n",Version);
         fprintf(stderr,"Usage: %s [{-d,-t <term>}] < in > out\n",Program);
         fprintf(stderr,"  -d  deletes all emphases\n"
                        "  -t  produces output suitable for <term> terminal or printer\n");
         exit(1);
      }
   }

#ifndef WITHOUT_CURSES
   if(OptimizeType==TERM)
   {
      if(!strcmp(term,"html"))
         html=1;

      setupterm(term,fileno(out),NULL);
/*      tputs(init_2string,1,MyPutc);
      tputs(init_3string,1,MyPutc);*/
      tputs(enter_ca_mode,1,MyPutc);

      Optimize();

      NewAttr=0;
      SetAttr();
      
      tputs(exit_ca_mode,1,MyPutc);
   }
   else
#endif
   {
      Optimize();
   }
   
   return(0);
}
